#!/usr/bin/env python3
# -*- conding:utf-8 -*-

from __future__ import absolute_import
import queue
import threading
import serial
from logging.handlers import TimedRotatingFileHandler
from datetime import datetime
import logging
from .constantes import *
from qgis.PyQt import QtCore
from builtins import str
from future import standard_library
standard_library.install_aliases()
import time
from binascii import unhexlify
import os

simu = False
simuSpare = False
simuLight = False
if 'SIMU' in os.environ:
    import random
    if int(os.environ['SIMU']) == 1:
        simu = True
    if int(os.environ['SIMU']) == 2:
        simuSpare = True
    if int(os.environ['SIMU']) == 3:
        simuLight = True
    if int(os.environ['SIMU']) == 4:
        simGammaFreeze = True

class rtkubeSerial(QtCore.QObject):

    sendGNSS = QtCore.pyqtSignal(str,  # DATE
                                 str,  # TIME
                                 str,  # LON
                                 str,  # LAT
                                 str,  # ALT
                                 str,  # PREC PLANI
                                 str,  # PREC ALTI
                                 str  # Calculation
                                 )

    sendGammaInc = QtCore.pyqtSignal(str,  # UpDGam
                                     str,  # LowDGam
                                     str  # Incli
                                     )

    sendError = QtCore.pyqtSignal(str)

    settings = QtCore.QSettings()

    def __init__(self):
        super(rtkubeSerial, self).__init__()
        self.ser = serial.Serial()
        self.running = False
        self.thread = None
        self.thread_stop_evt = None

        self.logger = logging.getLogger('rtkubeSerial')
        self.logger.setLevel(logging.DEBUG)
        fh = logging.handlers.TimedRotatingFileHandler(LOGFILE)
        fh.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)
        self.nbError = 0

    def connect(self):
        normalConnection = False
        error = ""
        target = None
        if simu:
            target= self.workerThreadSIMU
            print( 'Simulation mode')
        elif simuSpare:
            target = self.workerThreadBackupSIMUSpare
            print( 'Simulation spare mode')
        elif simuLight:
            target = self.workerThreadSIMULight
            print( 'Simulation light mode')
        elif simGammaFreeze:
            target = self.workerThreadSIMUGammaFreeze
        else:
            try:
                self.ser = serial.Serial(
                    self.settings.value('/GeoCube/serialportrover', ''), baudrate=BAUDRATE, timeout=TIMEOUT)
                normalConnection = True
                target = self.workerThread
            except serial.serialutil.SerialException as e:
                error = str(e)

            if not normalConnection:
                try:
                    self.ser = serial.Serial(self.settings.value('/GeoCube/serialportspare', ''), baudrate=BAUDRATE_BACKUP,
                                        timeout=TIMEOUT)
                    target = self.workerThreadBackup
                except serial.serialutil.SerialException as e:
                    self.disconnect()
                    self.logger.critical(e)
                    self.logger.critical(error)
                    self.sendError.emit("Cannot connect to bluetooth device nor spare : {} {}".format(e, error))

        if target:
            self.running = True
            if not simu and not simuSpare and not simuLight and not simGammaFreeze:
                self.ser.reset_input_buffer()
            self.thread_stop_evt = threading.Event()
            self.thread = threading.Thread(target=target,
                                           args=(self.thread_stop_evt,)).start()

    def disconnect(self):
        if self.thread_stop_evt:
            self.thread_stop_evt.set()
        self.running = False
        self.ser.close()

    def workerThread(self, stop_evt):
        while not stop_evt.is_set() and self.running and self.ser.isOpen():
            msg = None
            try:
                msg = self.ser.readline().decode('ascii')
            except serial.SerialException as e:
                self.disconnect()
                self.logger.error(e)
                self.sendError.emit(str(e))
            except Exception as e:
                self.disconnect()
                self.logger.error(e)
                self.sendError.emit("Unknown error: {}".format(str(e)))

            if msg and type(msg) == str and len(msg) > 0:
                self.logger.info(msg)
                fields = msg.split(FIELD_SEPARATOR)
                if len(fields) > MINIMAL_NB_FIELDS:
                    self.nbError = 0
                    if (fields[SENSOR] == SENSORS['GNSS'] and
                            len(fields) == NBFIELDS['GNSS']):
                        self.sendGNSS.emit(*fields[MINIMAL_NB_FIELDS:])
                    elif (fields[SENSOR] == SENSORS['GAMMA'] and
                          len(fields) == NBFIELDS['GAMMA']):
                        self.sendGammaInc.emit(
                            fields[UpDGam], fields[LowDGam], fields[Inc])
                else:
                    self.nbError += 1
                    self.disconnect()
                    self.logger.error("Incorrect value: {}\nVerify your com port.".format(msg))
                    self.sendError.emit("Incorrect value: {}\nVerify your com port.".format(msg))

            else:
                self.nbError += 1
                if self.nbError == 3:
                    self.disconnect()
                    msg = "No data received since 2 seconds."
                    self.logger.error(msg)
                    self.sendError.emit(msg)

    def getDataBackup(self):
        ret = []
        string = "054449" # It's kind of magic
        # 0x05 : ENQ
        # 0x44 : Gimme data
        # 0x49 : Checksum = 0x05+0x44
        self.ser.write(unhexlify(string))

        ack_nak = self.ser.read(1)
        ret.append(ack_nak) # ack_nak
        if ack_nak == b'\x06': # Ack : 0x06
            ret.append(self.ser.read(2)) # ggdm_serial
            ret.append(self.ser.read(3)) # upper_delta_count
            ret.append(self.ser.read(2)) # upper_delta_time
            ret.append(self.ser.read(4)) # upper_kfactor
            ret.append(self.ser.read(3)) # lower_delta_count
            ret.append(self.ser.read(2)) # lower_delta_time
            ret.append(self.ser.read(4)) # lower_kfactor
            ret.append(self.ser.read(2)) # inclination
            ret.append(self.ser.read(1)) # csum

        return ret

    def workerThreadBackup(self, stop_evt):
        while not stop_evt.is_set() and self.running and self.ser.isOpen():
            utc = datetime.utcnow()
            d = utc.strftime("%d/%m/%Y")
            t = utc.strftime("%H:%M:%S")
            try:
                ret = self.getDataBackup()
                if len(ret) > 1:
                    ack_nak, ggdm_serial, upper_delta_count, upper_delta_time, upper_kfactor, lower_delta_count, lower_delta_time, lower_kfactor, inclination, csum = ret
                    self.logger.info("upper_delta_count: {}".format(sum(upper_delta_count)))
                    self.logger.info("lower_delta_count: {}".format(sum(lower_delta_count)))
                    self.logger.info("inclination: {}".format(sum(inclination)))
                    self.sendGNSS.emit(d, t, "-1", "-1", "-1", "-1", "-1", "0")
                    time.sleep(1)
                    self.sendGammaInc.emit(str(sum(upper_delta_count)),
                                           str(sum(lower_delta_count)),
                                           str(sum(inclination)))
                else:
                    self.logger.error("NAK")
                    self.sendError.emit("Cannot get data from Spare")
            except serial.SerialException as e:
                self.disconnect()
                self.logger.error(e)
                self.sendError.emit(str(e))

    def workerThreadSIMU(self, stop_evt):
        while not stop_evt.is_set() and self.running:
            utc = datetime.utcnow()
            d = utc.strftime("%d/%m/%Y")
            t = utc.strftime("%H:%M:%S")
            lon = str(round(random.uniform(18.77984876, 18.77986872), 8))
            lat = str(round(random.uniform(7.33948401, 7.33950260), 8))

            self.sendGNSS.emit(d, t, lat, lon, "666", "-1", "-1", "23")
            time.sleep(1)
            self.sendGammaInc.emit(str(random.randrange(5, 1500)),
                                   str(random.randrange(5,1500)),
                                   str(random.randint(0, 10)))

    def workerThreadBackupSIMUSpare(self, stop_evt):
        while not stop_evt.is_set() and self.running:
            utc = datetime.utcnow()
            d = utc.strftime("%d/%m/%Y")
            t = utc.strftime("%H:%M:%S")
            self.sendGNSS.emit(d, t, "-1", "-1", "-1", "-1", "-1", "0")
            time.sleep(1)
            self.sendGammaInc.emit(str(random.randrange(5, 1500)),
                                   str(random.randrange(5,1500)),
                                   str(random.randint(0, 10)))

    def workerThreadSIMULight(self, stop_evt):
        while not stop_evt.is_set() and self.running:
            utc = datetime.utcnow()
            d = utc.strftime("%d/%m/%Y")
            t = utc.strftime("%H:%M:%S")
            lon = str(round(random.uniform(18.77984876, 18.77986872), 8))
            lat = str(round(random.uniform(7.33948401, 7.33950260), 8))

            self.sendGNSS.emit(d, t, lat, lon, "666", "-1", "-1", "23")
            time.sleep(1)
            self.sendGammaInc.emit("-1",
                                   "-1",
                                   "-1")

    def workerThreadSIMUGammaFreeze(self, stop_evt):
        last_time = time.time()
        count = 0
        while not stop_evt.is_set() and self.running:
            utc = datetime.utcnow()
            d = utc.strftime("%d/%m/%Y")
            t = utc.strftime("%H:%M:%S")
            lon = str(round(random.uniform(18.77984876, 18.77986872), 8))
            lat = str(round(random.uniform(7.33948401, 7.33950260), 8))

            self.sendGNSS.emit(d, t, lat, lon, "666", "-1", "-1", "23")
            time.sleep(1)
            if last_time < time.time() - 10 :
                self.sendGammaInc.emit(str(600),
                                   str(300),
                                   str(random.randint(0, 10)))
                count += 1
                if count > 5 :
                    last_time = time.time()
                    count = 0
            else :
                self.sendGammaInc.emit(str(random.randrange(5, 1500)),
                                   str(random.randrange(5,1500)),
                                   str(random.randint(0, 10)))
